/* Name: main.c
 * Project: Programming Pen - USB keyboard interface
 * Author: Anthony Van Herrewege
 * Date: 2013-04-24
 *
 * Based on EasyLogger code by Christian Starkjohann.
 * See http://www.obdev.at/avrusb/easylogger.html.
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "usbdrv.h"
#include "oddebug.h"

/*
 * Pin assignment:
 * PB1 = key input (active low with pull-up)
 * PB3 = analog input (ADC3)
 * PB4 = LED output (active high)
 *
 * PB0, PB2 = USB data lines
 */

#define BIT_LED 4
#define BIT_KEY 3

#ifndef NULL
  #define NULL    ((void *)0)
#endif

/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */

static uchar    isTransmitting;

/* ------------------------------------------------------------------------- */

const PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = { /* USB report descriptor */
	0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
	0x09, 0x06,                    // USAGE (Keyboard)
	0xa1, 0x01,                    // COLLECTION (Application)
	0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
	0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
	0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
	0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
	0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
	0x75, 0x01,                    //   REPORT_SIZE (1)
	0x95, 0x08,                    //   REPORT_COUNT (8)
	0x81, 0x02,                    //   INPUT (Data,Var,Abs)
	0x95, 0x01,                    //   REPORT_COUNT (1)
	0x75, 0x08,                    //   REPORT_SIZE (8)
	0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
	0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
	0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
	0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
	0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */

#define KEY_A       4
#define KEY_B       5
#define KEY_C       6
#define KEY_D       7
#define KEY_E       8
#define KEY_F       9
#define KEY_G       10
#define KEY_H       11
#define KEY_I       12
#define KEY_J       13
#define KEY_K       14
#define KEY_L       15
#define KEY_M       16
#define KEY_N       17
#define KEY_O       18
#define KEY_P       19
#define KEY_Q       20
#define KEY_R       21
#define KEY_S       22
#define KEY_T       23
#define KEY_U       24
#define KEY_V       25
#define KEY_W       26
#define KEY_X       27
#define KEY_Y       28
#define KEY_Z       29
#define KEY_1       30
#define KEY_2       31
#define KEY_3       32
#define KEY_4       33
#define KEY_5       34
#define KEY_6       35
#define KEY_7       36
#define KEY_8       37
#define KEY_9       38
#define KEY_0       39

#define KEY_RETURN  40
#define KEY_SPACE   44

#define KEY_UP      82

/* ------------------------------------------------------------------------- */

static void buildReport (void)
{
	if (reportBuffer[1] == 0)
		reportBuffer[1] = KEY_UP;
	else if (reportBuffer[1] == KEY_UP)
		reportBuffer[1] = KEY_RETURN;
	else
		reportBuffer[1] = 0;
}

/* ------------------------------------------------------------------------- */

static void setIsTransmitting (uchar newValue)
{
	isTransmitting = newValue;

	if (isTransmitting) PORTB &= ~(1 << BIT_LED); /* LED off */
	else								PORTB |= 1 << BIT_LED;    /* LED on */
}

/* ------------------------------------------------------------------------- */

static void keyPoll(void)
{
	static uchar    keyMirror;
	uchar           key;

	key = PINB & (1 << BIT_KEY);
	if (keyMirror != key)
	{   // status changed
		keyMirror = key;
		// Key was pressed
		if (!key) setIsTransmitting(1);
	}
}

static void timerPoll(void)
{
	static uchar timerCnt;

	if (TIFR & (1 << TOV1))
	{
		TIFR = (1 << TOV1); /* clear overflow */
		keyPoll();
		if (++timerCnt >= 63)
		{       /* ~ 1 second interval */
			timerCnt = 0;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void timerInit(void)
{
	TCCR1 = 0x0b;           /* select clock: 16.5M/1k -> overflow rate = 16.5M/256k = 62.94 Hz */
}

/* ------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* ------------------------------------------------------------------------- */

uchar	usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (void *)data;

	usbMsgPtr = reportBuffer;
	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
	{    /* class request type */
		if (rq->bRequest == USBRQ_HID_GET_REPORT)
		{  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
			/* we only have one report type, so don't look at wValue */
			//buildReport();
			return sizeof(reportBuffer);
		} else if (rq->bRequest == USBRQ_HID_GET_IDLE)
		{
			usbMsgPtr = &idleRate;
			return 1;
		} else if (rq->bRequest == USBRQ_HID_SET_IDLE)
		{
			idleRate = rq->wValue.bytes[1];
		}
	} else
	{
		/* no vendor specific requests implemented */
	}

	return 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

#define abs(x) ((x) > 0 ? (x) : (-x))

// Called by V-USB after device reset
void hadUsbReset()
{
	int frameLength, targetLength = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);
	int bestDeviation = 9999;
	uchar trialCal, bestCal = 64, step, region;

	// do a binary search in regions 0-127 and 128-255 to get optimum OSCCAL
	for (region = 0; region <= 1; region++)
	{
		frameLength = 0;
		trialCal = (region == 0) ? 0 : 128;

		for (step = 64; step > 0; step >>= 1)
		{
			if(frameLength < targetLength) // true for initial iteration
				trialCal += step; // frequency too low
			else
				trialCal -= step; // frequency too high

			OSCCAL = trialCal;
			frameLength = usbMeasureFrameLength();

			if (abs(frameLength-targetLength) < bestDeviation)
			{
				bestCal = trialCal; // new optimum found
				bestDeviation = abs(frameLength -targetLength);
			}
		}
	}

	OSCCAL = bestCal;
}

/* ------------------------------------------------------------------------- */
/* --------------------------------- main ---------------------------------- */
/* ------------------------------------------------------------------------- */

int main(void)
{
	uchar   i;

	odDebugInit();
	usbDeviceDisconnect();

	for(i=0;i<20;i++){  /* 300 ms disconnect */
		_delay_ms(15);
	}
	usbDeviceConnect();

	// Port setup
	DDRB |= 1 << BIT_LED;   /* output for LED */
	PORTB |= 1 << BIT_KEY;  /* pull-up on key input */

	// Variable setup
	reportBuffer[0] = 0; // No modifiers
	reportBuffer[1] = 0; // No key
	setIsTransmitting(0);

	wdt_enable(WDTO_1S);
	//wdt_disable();

	timerInit();
	usbInit();
	sei();

  // Main event loop
	for (;;)
	{
		wdt_reset();
		usbPoll();

		if (usbInterruptIsReady() && isTransmitting)
		{ /* we can send another key */
			buildReport();
			usbSetInterrupt(reportBuffer, sizeof(reportBuffer));

			if (reportBuffer[1] == 0) setIsTransmitting(0);
		}

		timerPoll();
	}

	return 0;
}
